import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

import 'package:random/cubit/game_cubit.dart';
import 'package:random/models/game/game.dart';
import 'package:random/models/game/game_settings.dart';
import 'package:random/ui/painters/cell_painter.dart';
import 'package:random/ui/widgets/game/game_score.dart';

class GameBoardWidget extends StatefulWidget {
  const GameBoardWidget({
    super.key,
    required this.game,
    required this.widgetSize,
  });

  final Game game;
  final Size widgetSize;

  @override
  State<GameBoardWidget> createState() => _GameBoardWidget();
}

class _GameBoardWidget extends State<GameBoardWidget> with TickerProviderStateMixin {
  List<List<Animation<double>?>> animations = [];

  void resetAnimations(GameSettings gameSettings) {
    final boardSize = gameSettings.boardSize;

    animations = List.generate(
      boardSize,
      (i) => List.generate(
        boardSize,
        (i) => null,
      ),
    );
  }

  void removeCell(BuildContext context, int x, int y) {
    final GameCubit gameCubit = BlocProvider.of<GameCubit>(context);
    final Game updatedGame = gameCubit.state.game ?? Game.createNew();

    // "remove" cell, update counters
    updatedGame.increaseScore(updatedGame.getCellValue(x, y));
    updatedGame.increaseMovesCount();
    updatedGame.updateCellValue(x, y, null);
    setState(() {});

    // "move down" cells
    final controller = AnimationController(
      vsync: this,
      duration: Duration(milliseconds: 750),
    )..addListener(() {
        if (mounted) {
          setState(() {});
        }
      });

    if (mounted) {
      setState(() {});
    }

    Animation<double> animation = Tween(
      begin: 0.0,
      end: 1.0,
    ).animate(CurvedAnimation(
      curve: Curves.bounceOut,
      parent: controller,
    ))
      ..addStatusListener((status) {
        if (status == AnimationStatus.completed) {
          // Update cell values
          for (var i = 0; i < y; i++) {
            updatedGame.updateCellValue(x, (y - i), updatedGame.getCellValue(x, (y - i) - 1));
          }
          updatedGame.setRandomCellValue(x, 0, updatedGame.settings);

          resetAnimations(updatedGame.settings);
          setState(() {});

          controller.dispose();
        }
      });

    for (var i = 0; i < y; i++) {
      animations[(y - i) - 1][x] = animation;
    }

    controller.forward().orCancel;
  }

  Widget buildBoard() {
    final widgetWidth = widget.widgetSize.width;
    final widgetHeight = widget.widgetSize.height;

    final rowsCount = widget.game.settings.boardSize;
    final columnsCount = widget.game.settings.boardSize;

    final cellWidth = widgetWidth / columnsCount;
    final cellHeight = widgetHeight / rowsCount;

    if (animations.length == 0) {
      resetAnimations(widget.game.settings);
    }

    final List<Widget> cells = [];

    for (var y = 0; y < rowsCount; y++) {
      for (var x = 0; x < columnsCount; x++) {
        final int? value = widget.game.getCellValue(x, y);

        if (value != null) {
          final Animation<double>? translation = this.animations[y][x];

          final Widget cellContent = CustomPaint(
            size: Size(cellWidth, cellHeight),
            willChange: false,
            painter: CellPainter(value: value),
          );

          final Widget cellWidget = Positioned(
            left: (x * cellWidth).toDouble(),
            top: ((y + (translation?.value ?? 0)) * cellHeight).toDouble(),
            child: Container(
              width: cellWidth,
              height: cellHeight,
              child: cellContent,
            ),
          );

          cells.add(cellWidget);
        }
      }
    }

    return Container(
      width: widgetWidth,
      height: widgetHeight,
      color: Colors.black,
      child: Stack(
        children: cells,
      ),
    );
  }

  Widget interactiveBoard(BuildContext context) {
    final widgetWidth = widget.widgetSize.width;
    final widgetHeight = widget.widgetSize.height;

    final rowsCount = widget.game.settings.boardSize;
    final columnsCount = widget.game.settings.boardSize;

    return GestureDetector(
      child: buildBoard(),
      onTapUp: (details) {
        double xTap = details.localPosition.dx;
        double yTap = details.localPosition.dy;

        int x = (xTap / widgetWidth * columnsCount).toInt();
        int y = (yTap / widgetHeight * rowsCount).toInt();
        print('[' + x.toString() + ',' + y.toString() + ']');

        removeCell(context, x, y);
      },
    );
  }

  @override
  Widget build(BuildContext context) {
    return Column(
      children: [
        interactiveBoard(context),
        GameScoreWidget(game: widget.game),
      ],
    );
  }
}
